# file: rfcomm-server.py
# auth: Albert Huang <albert@csail.mit.edu>
# desc: simple demonstration of a server application that uses RFCOMM sockets
#
# $Id: rfcomm-server.py,v 1.7 2006/05/06 00:36:52 albert Exp $

from bluetooth import *
import location
import gobject
import time

#---------------------------------------------------------------------------------------
# Location defines
def on_error(control, error, data):
	print "location error: %d... quitting" % error
	data.quit()

def on_changed(device, data):
    if not device:
        return
    if device.fix:
        if device.fix[1] & location.GPS_DEVICE_LATLONG_SET:
			print device.fix
			process_lr_update(device, data)

def on_stop(control, data):
	print "quitting"
    data.quit()

def start_location(data):
    data.start()
    return False

#---------------------------------------------------------------------------------------
# Bluetooth defines
# Generate the checksum for some data
# (Checksum is all the data XOR'd, then turned into hex)
def generate_checksum(data):
	"""Generate the NMEA checksum for the supplied data"""
	csum = 0
	for c in data:
		csum = csum ^ ord(c)
	hex_csum = "%02x" % csum
	return hex_csum.upper()

# Format a NMEA timestamp into something friendly
def format_time(time):
	"""Generate a friendly form of an NMEA timestamp"""
	hh = time[0:2]
	mm = time[2:4]
	ss = time[4:]
	return "%s:%s:%s UTC" % (hh,mm,ss)

# Format a NMEA date into something friendly
def format_date(date):
	"""Generate a friendly form of an NMEA date"""
	dd = int(date[0:2])
	mm = int(date[2:4])
	yy = int(date[4:6])
	yyyy = yy + 2000
	return format_date_from_parts(yyyy,mm,dd)

def format_date_from_parts(yyyy,mm,dd):
	"""Generate a friendly date from yyyy,mm,dd"""
	months = ('Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec')
	return "%02d %s %d" % (dd, months[(int(mm)-1)], yyyy)

def nmea_format_latlong(lat,long):
	"""Turn lat + long into nmea format"""

	def format_ll(hour_digits,ll):
		abs_ll = abs(ll)
		hour_ll = int(abs_ll)
		min_ll = (abs_ll - hour_ll) * 60
		# 4dp, so *10,000
		str_ll = "%03d%02d.%04d" % ( hour_ll, int(min_ll), 10000*(min_ll-int(min_ll)) )
		if hour_digits == 2 and str_ll[0] == '0':
			str_ll = str_ll[1:]
		return str_ll

	if str(lat) == 'NaN':
		flat = '0000.000,N'
	else:
		flat = format_ll(2, lat) + ","
		if lat < 0:
			flat += "S"
		else:
			flat += "N"

	if str(long) == 'NaN':
		flong = '00000.000,E'
	else:
		flong = format_ll(3, long) + ","
		if long < 0:
			flong += "W"
		else:
			flong += "E"
	return (flat,flong)

def user_format_latlong(lat,long):
	"""Turn lat + long into a user facing format"""
	if str(lat) == 'NaN':
		flat = '00:00:00N'
	else:
		alat = abs(lat)
		mins = (alat-int(alat))*60
		secs = (mins-int(mins))*60
		flat = "%02d:%02d:%02d" % (int(alat),int(mins),int(secs))
		if lat < 0:
			flat += "S"
		else:
			flat += "N"
	if str(long) == 'NaN':
		flong = '000:00:00W'
	else:
		along = abs(long)
		mins = (along-int(along))*60
		secs = (mins-int(mins))*60
		flong = "%03d:%02d:%02d" % (int(along),int(mins),int(secs))
		if long < 0:
			flong += "W"
		else:
			flong += "E"
	return (flat,flong)

# NMEA data is HHMM.nnnn where nnnn is decimal part of second
def format_latlong(data):
	"""Turn HHMM.nnnn into HH:MM.SS"""

	# Check to see if it's HMM.nnnn or HHMM.nnnn or HHHMM.nnnn
	if data[5:6] == '.':
		# It's HHHMM.nnnn
		hh_mm = data[0:3] + ":" + data[3:5]
		dddd = data[6:]
	elif data[3:4] == '.':
		# It's HMM.nnnn
		hh_mm = data[0:1] + ":" + data[1:3]
		dddd = data[4:]
	else:
		# Assume HHMM.nnnn
		hh_mm = data[0:2] + ":" + data[2:4]
		dddd = data[5:]

	# Turn from decimal into seconds, and strip off last 2 digits
	sec = int( float(dddd) / 100.0 * 60.0 / 100.0 )
	return hh_mm + ":" + str(sec)

def format_latlong_dec(data):
	"""Turn HHMM.nnnn into HH.ddddd"""
	
	# Check to see if it's HMM.nnnn or HHMM.nnnn or HHHMM.nnnn
	if data[5:6] == '.':
		hours = data[0:3]
		mins = float(data[3:])
	elif data[3:4] == '.':
		hours = data[0:1]
		mins = float(data[1:])
	else:
		hours = data[0:2]
		mins = float(data[2:])

	dec = mins / 60.0 * 100.0
	# Cap at 6 digits - currently nn.nnnnnnnn
	dec = dec * 10000.0
	str_dec = "%06d" % dec
	return hours + "." + str_dec

def get_latlong_floats():
	global locationDict
	wgs_lat = locationDict['lat_dec'];
	wgs_long = locationDict['long_dec'];
	if wgs_lat[-1:] == 'S':
		wgs_lat = '-' + wgs_lat;
	if wgs_long[-1:] == 'W':
		wgs_long = '-' + wgs_long;
	wgs_lat = float(wgs_lat[0:-1])
	wgs_long = float(wgs_long[0:-1])

	return (wgs_lat,wgs_long)

#---------------------------------------------------------------------------------------

sock = None
def process_lr_update(device, data):
	"""Process a locationDict update from locationDictRequestor"""
	global satellites
	global locationDict
	global motion
	global gps
	global sock
	#print args
	
	if not device:
		return
	
	# was if args > 2, need to see what we do abou this
	if True:
		
		locationDict['lat']  = device.fix[4]
		locationDict['long']  = device.fix[5]

		if str(locationDict['lat']) == 'NaN':
			locationDict['lat_dec'] = "0.0"
		else:
			locationDict['lat_dec'] = "%02.6f" % locationDict['lat']
		if str(locationDict['long']) == 'NaN':
			locationDict['long_dec'] = "0.0"
		else:
			locationDict['long_dec'] = str(locationDict['long'])

		locationDict['alt'] = "%3.1f m" % (device.fix[7])

		satellites['horiz_dop'] = "%0.1f" % device.fix[6]
		satellites['vert_dop'] = "%0.1f" % device.fix[8]
		satellites['overall_dop'] = "%0.1f" % ((device.fix[6]+device.fix[8])/2)
		
		#if len(args) > 8:
		if(str(device.fix[11])) == 'NaN':
			if motion.has_key('speed'):
				del motion['speed']
		else:
			motion['speed_kmph'] = float(device.fix[11])
			motion['speed_mph'] = float(device.fix[11]) / 1.609344
		if(str(device.fix[9])) == 'NaN':
			if motion.has_key('true_heading'):
				del motion['true_heading']
		else:
			motion['true_heading'] = "%0.1f" % device.fix[9]

		locationDict['tsecs'] = long(device.fix[2]/1000) # ms not sec
		timeparts = time.gmtime( locationDict['tsecs'] )
		locationDict['time'] = "%02d:%02d:%02d" % (timeparts[3:6])
		locationDict['date'] = format_date_from_parts(*timeparts[0:3])
		gga_time = "%02d%02d%02d.%02d" % (timeparts[3:7])

		if device.satellites_in_use == 0:
			locationDict['valid'] = 0
		else:
			locationDict['valid'] = 1
	else:
		if str(device.fix[4]) == 'NaN':
			locationDict['valid'] = 0
		else:
			locationDict['valid'] = 1

		locationDict['tsecs'] = long(device.fix[2]/1000) # ms not sec
		timeparts = time.gmtime( locationDict['tsecs'] )
		locationDict['time'] = "%02d:%02d:%02d" % (timeparts[3:6])
		locationDict['date'] = format_date_from_parts(*timeparts[0:3])
		gga_time = "%02d%02d%02d.%02d" % (timeparts[3:7])


	# Figure out the satellites in view
	in_view = []
	in_use = []
	
	# Send NMEA data over the socket
	if sock:
		# GGA - position
		nmea_latlong = nmea_format_latlong(device.fix[4],device.fix[5])
		checkSum = generate_checksum("GPGGA,%s,%s,%s,%d,%02d,0.0,%0.2f,M,,,," % gga_time,nmea_latlong[0],nmea_latlong[1],locationDict['valid'],
					len(in_view),device.fix[7])
		sock.send(
			"$GPGGA,%s,%s,%s,%d,%02d,0.0,%0.2f,M,,,,*%s\n" % \
				(gga_time,nmea_latlong[0],nmea_latlong[1],locationDict['valid'],
					len(in_view),device.fix[7], checkSum)
		)
		print "Sending to sock : $GPGGA,%s,%s,%s,%d,%02d,0.0,%0.2f,M,,,,*%s\n" % (gga_time,nmea_latlong[0],nmea_latlong[1],locationDict['valid'], len(in_view),device.fix[7], checkSum)
		
		# VTG - track made good (motion) 
		heading = ""
		speed = ""
		if motion.has_key('true_heading'):
			heading = motion['true_heading']
		if motion.has_key('speed_kmph'):
			speed = "%3.1f" % motion['speed_kmph']
		checkSum = generate_checksum("GPVTG,%s,T,,M,%s,N,%s,K,A" % heading,"000.0",speed)
		sock.send(
				"$GPVTG,%s,T,,M,%s,N,%s,K,A*%s\n" % (
					heading,"000.0",speed, checkSum
				)
		)
		print "Sending to sock : $GPVTG,%s,T,,M,%s,N,%s,K,A*%s\n" % (heading,"000.0",speed, checkSum)

#---------------------------------------------------------------------------------------
# This is set to 0 to request a quit
going = 1
# Our current locationDict
locationDict = {}
locationDict['valid'] = 1 # Default to valid, in case no GGA/GLL sentences
# Our current motion
motion = {}
# What satellites we're seeing
satellites = {}
# Warnings / errors
disp_notices = ''
disp_notices_count = 0

server_sock=BluetoothSocket( RFCOMM )
server_sock.bind(("",PORT_ANY))
server_sock.listen(1)

port = server_sock.getsockname()[1]

uuid = "94f39d29-7d6d-437d-973b-fba39e49d4ee"

advertise_service( server_sock, "Serial",
                   service_id = uuid,
                   service_classes = [ uuid, SERIAL_PORT_CLASS ],
                   profiles = [ SERIAL_PORT_PROFILE ] )
                   
print "Waiting for connection on RFCOMM channel %d" % port

client_sock, client_info = server_sock.accept()
sock = client_sock
print "Accepted connection from ", client_info

loop = gobject.MainLoop()
control = location.GPSDControl.get_default()
device = location.GPSDevice()
control.set_properties(preferred_method=location.METHOD_USER_SELECTED,
                       preferred_interval=location.INTERVAL_DEFAULT)

control.connect("error-verbose", on_error, loop)
device.connect("changed", on_changed, control)
control.connect("gpsd-stopped", on_stop, loop)

gobject.idle_add(start_location, control)

loop.run()

print "disconnected"

client_sock.close()
server_sock.close()